/***************************************************************************
                          qlabyrinthdoc.cpp  -  description
                             -------------------
    begin                : Die 19 Jun 2001 19:28:20 MEST
    copyright            : (C) 2001 by Ruediger Knoerig
    email                : ruediger@knoerig.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include <qpoint.h>
#include <qsize.h>
#include <qstring.h>
#include <qrect.h>
#include <iostream.h>
#include <fstream.h>
#include "qlabyrinthdoc.h"

extern "C"
{
#include <string.h>
}

QLabyrinthDoc::QLabyrinthDoc(const QSize &LabySize)
{
	dimensions = LabySize;
	minlabycelldim = 3;
	
	if((laby = new fieldstate[dimensions.width()*dimensions.height()]) == 0) throw QString("not enough memory for the labyrinth!");	
	
	memset(laby,0,dimensions.width()*dimensions.height()*sizeof(fieldstate));
	makeLaby();
	emit newStepCount(stepcount);
	
  modified = false;
}

QLabyrinthDoc::~QLabyrinthDoc()
{
	if(laby) delete[] laby;
}

void QLabyrinthDoc::newDoc(const QSize &LabySize)
{
	// lock access to labyrinth data
	while(!datamutex.lock());

	dimensions = LabySize;
			
	delete laby;
	if((laby = new fieldstate[dimensions.width()*dimensions.height()]) == 0) throw QString("not enough memory for the labyrinth!");	
	memset(laby,0,dimensions.width()*dimensions.height()*sizeof(fieldstate));
	
	// clear movement stack
	while(!moves.empty()) moves.pop();
	
	makeLaby();
	
	// unlock access to labyrinth data
	datamutex.unlock();
	
	emit documentChanged();
	emit newStepCount(stepcount);
}

bool QLabyrinthDoc::save()
{
	if(!filename.isEmpty())
	{
		saveAs(filename);
	}
  return true;
}

bool QLabyrinthDoc::saveAs(const QString &filename)
{
	this->filename = filename;

	ofstream output((const char *) filename);
	if(!output) throw (QString("Cannot open ") + filename + QString(" !"));
	
	header hdr;
	hdr.id = *((unsigned long int *) idstring);
	hdr.stepcounter = stepcount;
	hdr.cookieorient = (char) act_dir;
	hdr.cookieframe = (char) cookiestep;
	hdr.cookiex = (char) cookiepos.x();
	hdr.cookiey = (char) cookiepos.y();
	hdr.width = (char) dimensions.width();
	hdr.height = (char) dimensions.height();
	
	// write header
	output.write(&hdr,sizeof(hdr));

	// copy laby data to char array for saving disk space
	int size = dimensions.width()*dimensions.height();
	char *cpy = new char[size];
	char *dptr = cpy;
	QLabyrinthDoc::fieldstate *ptr = &laby[0];
	for(int i=0; i < size; i++,ptr++,dptr++) *dptr = (char) *ptr;
	
	// write data	
	output.write(cpy,size*sizeof(char));
	delete[] cpy;
	
	// write move stack
	stack<char> temp(moves);
	while(!temp.empty())
	{
		output.put(temp.top());
		temp.pop();
	}
	
	if(!output)
	{
		output.close();
		throw QString("Error while saving labyrinth data!");
	}
	
	output.close();
  return true;
}

bool QLabyrinthDoc::load(const QString &filename)
{
	ifstream input((const char *) filename);
	if(!input) throw (QString("Cannot open ") + filename + QString(" !"));
	
	
	
	header hdr;
	hdr.id = 0;
	
	input.read(&hdr,sizeof(hdr));
	if(!input || (hdr.id != *((unsigned long int *) idstring)))
	{
		input.close();
		throw QString(filename + QString("\nseems not to be a valid\nQLabyrinth data file!"));
	}
	
	
	emit newStepCount(stepcount = (int) hdr.stepcounter);
	
	// lock access to labyrinth data
	while(!datamutex.lock());
	
	act_dir = (QLabyrinthDoc::orientation) hdr.cookieorient;
	cookiestep = hdr.cookieframe;
	
	// calculate number of desired array elements
	
	int size = int(hdr.width) * int(hdr.height);
		
	// copy laby data to char array for saving disk space
	
	delete laby;
	
	if(!(laby = new fieldstate[size]))
	{
		datamutex.unlock();
		throw QString("not enough memory for the labyrinth!");
	}	
	
	memset(laby,0,size*sizeof(fieldstate));
	
	dimensions = QSize(int(hdr.width),int(hdr.height));
	
	char *cpy = new char[size];
	char *dptr = cpy;
	
	// read data
	input.read(&cpy[0],size*sizeof(char));
	
	// clear old movement stack
	while(!moves.empty()) moves.pop();
	// read movement stack
	while(input && !input.eof())
		{
			 moves.push((char) input.get());
		}
	moves.pop();
	cookiepos = QPoint(hdr.cookiex,hdr.cookiey);
	cookieptr = (laby + hdr.cookiex + int(hdr.cookiey) * int(hdr.width));
		
	// extract some data from the array
	QLabyrinthDoc::fieldstate *ptr = &laby[0];
	for(int y=0; y < (int) hdr.height; y++)
		for(int x=0; x < (int) hdr.width; x++, ptr++,dptr++)
			{
				*ptr = (QLabyrinthDoc::fieldstate) *dptr;
				switch(*ptr)
				{
					case entry : entrypos = QPoint(x,y);
						break;
					case exit : exitpos = QPoint(x,y);
						break;
					default:
						break;
				}
			}
	datamutex.unlock();

  emit documentChanged();
	
	delete[] cpy;
  return true;
}

bool QLabyrinthDoc::isModified() const
{
  return modified;
}

/** generates a labyrinth */
void QLabyrinthDoc::makeLaby()
{
	QLabyrinthDoc::fieldstate *ptr = &laby[0];

	int w = dimensions.width();
	int h = dimensions.height();

	// clear laby
	memset(laby,0,sizeof(fieldstate)*w*h);
		
	// make the upper wall
	for(int i=0; i < w; i++,ptr++)
		*ptr = QLabyrinthDoc::wall;
	// make the side walls
	for(int i=1; i < h-1; i++)
		{
			*ptr = wall;
			ptr += (w-1);
			*(ptr++) = wall;
		}
	// make the lower wall
	for(int i=0; i < w; i++,ptr++) *ptr = wall;
	ptr++;
	// estaminate entry + cookie and exit coordinates
	int xpos = ((int) ((w - 2.0)*rand()/(RAND_MAX+1.0))) | 0x1;
	cookieptr = laby + xpos;
	cookiepos = entrypos = QPoint(xpos,0);
	laby[xpos] = entry;
	// using QPoints for entry & exit coordinates would give more freedom
	xpos = ((int) ((w - 2.0)*rand()/(RAND_MAX+1.0))) | 0x1;
	
	exitpos = QPoint(xpos,h-1);
	*(ptr - w + xpos - 1) = exit;
	
	// now lets make the inner labyrinth recursively
	rekursiveLaby(&laby[0],dimensions.width(),dimensions.height());
	
	act_dir = down;
	cookiestep = 0;
	stepcount = 0;
}


/** returns the fieldtype of the adjacent field in direction dir (and wall, if outside of the laby) */
QLabyrinthDoc::fieldstate QLabyrinthDoc::whatsAt(QLabyrinthDoc::orientation dir)
{
	switch(dir)
	{
		case up :
			if(cookiepos.y() < 1) return wall;
				else return *(cookieptr - dimensions.width());
			break;
		case down :
			if(cookiepos.y() < dimensions.height())
				return *(cookieptr + dimensions.width());
				else return wall;
			break;
		case left :
			if(cookiepos.x() < 1) return wall;
				else return *(cookieptr - 1);
			break;
		case right :
			if(cookiepos.x() < dimensions.width())
				return *(cookieptr + 1);
			else return wall;
			break;
		default : return wall; 	
	}
}

/**
	* moves cookie in direction dir _without_ checking for correctnes
	* (use whatsAt for checking)
 	*/
void QLabyrinthDoc::move(QLabyrinthDoc::orientation dir)
{
	QRect updatearea(cookiepos.x() << 5, cookiepos.y() << 5,32,64);
	
	cookiestep ^= 0x1;
	act_dir = dir;
	moves.push((char) dir);
	
	if(*cookieptr == QLabyrinthDoc::empty) *cookieptr = passed;
	else if(*cookieptr == QLabyrinthDoc::passed) *cookieptr = passed2;
	
	switch(dir)
		{
			case up :
				updatearea.moveBy(0,-32);
				cookieptr -= dimensions.width();
				cookiepos.setY(cookiepos.y()-1);
				break;
			case down :
				cookieptr += dimensions.width();
				cookiepos.setY(cookiepos.y()+1);
				break;
			case left :
				updatearea.moveBy(-32,0);
				updatearea.setWidth(64);
				updatearea.setHeight(32);
				cookieptr -= 1;
				cookiepos.setX(cookiepos.x()-1);
				break;
			case right :
				updatearea.setWidth(64);
				updatearea.setHeight(32);
				cookieptr += 1;
				cookiepos.setX(cookiepos.x()+1);
				break;
		}
		
	emit updateView(updatearea,false);
	emit newStepCount(++stepcount);
}

/** Erzeugt den Labyrinthinhalt nach einer rekursiven Prozedur
	* @param begin Zeiger auf die obere linke Ecke des Sublabyrinthes
	* @param width Breite des Sublabyrinthes
	* @param height Hhe des Sublabyrinthes
	 */
void QLabyrinthDoc::rekursiveLaby(fieldstate *begin, int width, int height)
{
	int max = ((width > height) ? width : height);
	
	if(max > minlabycelldim)
	{
		int offset = dimensions.width();
		if(width >= height)
			{
				// make a vertical wall at an even position
				int pos = (2 + (int) ((width - 4.0)*rand()/(RAND_MAX+1.0))) | 0x1;
				pos ^= 0x1;
				
				rekursiveLaby(begin,pos,height);
				begin += pos;
				rekursiveLaby(begin,width-pos-1,height);
				
				// draws the wall
			  fieldstate *ptr = begin + offset;		
				for(int i=1;i<height-1;i++,ptr+=offset) *ptr = wall;
				// mark the passage
				pos = (1 + (int) ((height - 3.0)*rand()/(RAND_MAX+1.0))) | 0x1;
				*(begin + pos * offset) = empty;
			}
			else
			{
				// make a horizontal wall at an even position
				int pos = (2 + (int) ((height - 4.0)*rand()/(RAND_MAX+1.0))) | 0x1;
				pos ^= 0x1;
				
				rekursiveLaby(begin,width,pos);
				begin += pos * offset;
				rekursiveLaby(begin,width,height-pos-1);
				// draw the wall
				fieldstate *ptr = begin + 1;
				for(int i=1;i<width-1;i++,ptr++) *ptr = wall;
				// marks the passage
				pos = (1 + (int) ((width - 3.0)*rand()/(RAND_MAX+1.0))) | 0x1;
				*(begin + pos) = empty;
			}
	}
	
}
/**
	* If there are protocolled moves, the last will be undone.
	* @return true if succeed, false if move stack is empty
	*/
bool QLabyrinthDoc::undoLastMove()
{
	if(moves.empty()) return false;
	
	QLabyrinthDoc::orientation ivstep;
	QLabyrinthDoc::orientation laststep = ((QLabyrinthDoc::orientation) moves.top());
	// remove the element
	moves.pop();
	// invert the move
	switch(laststep)
	{
		case up : ivstep = down;
		break;
		case down : ivstep = up;
		break;
		case left : ivstep = right;
		break;
		case right : ivstep = left;
		break;
	}
	
	bool success = (whatsAt(ivstep) != wall);
	if(success)
	{
		move(ivstep);
		// clear that move from the stack
		moves.pop();
	}
	return success;
}
